import { createUnplugin } from 'unplugin';
import { createRequire } from 'node:module';
import * as path from 'node:path';
import { join } from 'node:path';
import { createLanguageServiceHost, decorateLanguageService } from '@volar/typescript';
import * as vue from '@vue/language-core';
import MagicString from 'magic-string';
import ts from 'typescript';

const windowsPathReg = /\\/g;
const normalizePath = (id) => id.replace(windowsPathReg, "/");
const resolveOptions = (rawOptions) => ({
  tsconfigPath: rawOptions.tsconfigPath ?? join(process.cwd(), "tsconfig.json")
});

const require = createRequire(import.meta.url);
function createLanguage(tsconfigPath, ts = require("typescript")) {
  const tsconfig = normalizePath(tsconfigPath);
  return createLanguageWorker(
    () => vue.createParsedCommandLine(ts, ts.sys, tsconfigPath),
    path.dirname(tsconfig),
    ts
  );
}
function createLanguageWorker(loadParsedCommandLine, rootPath, ts) {
  let parsedCommandLine = loadParsedCommandLine();
  let fileNames = parsedCommandLine.fileNames.map(normalizePath);
  let projectVersion = 0;
  const scriptSnapshots = /* @__PURE__ */ new Map();
  const _host = {
    workspacePath: rootPath,
    rootPath,
    getProjectVersion: () => projectVersion.toString(),
    getCompilationSettings: () => parsedCommandLine.options,
    getScriptFileNames: () => fileNames,
    getProjectReferences: () => parsedCommandLine.projectReferences,
    getScriptSnapshot: (fileName) => {
      if (!scriptSnapshots.has(fileName)) {
        const fileText = ts.sys.readFile(fileName);
        if (fileText !== void 0) {
          scriptSnapshots.set(fileName, ts.ScriptSnapshot.fromString(fileText));
        }
      }
      return scriptSnapshots.get(fileName);
    }
  };
  return {
    ...baseCreateLanguageWorker(
      _host,
      vue.resolveVueCompilerOptions(parsedCommandLine.vueOptions),
      ts
    ),
    updateFile(fileName, text) {
      fileName = normalizePath(fileName);
      scriptSnapshots.set(fileName, ts.ScriptSnapshot.fromString(text));
      projectVersion++;
    },
    deleteFile(fileName) {
      fileName = normalizePath(fileName);
      fileNames = fileNames.filter((f) => f !== fileName);
      projectVersion++;
    },
    reload() {
      parsedCommandLine = loadParsedCommandLine();
      fileNames = parsedCommandLine.fileNames.map(normalizePath);
      this.clearCache();
    },
    clearCache() {
      scriptSnapshots.clear();
      projectVersion++;
    }
  };
}
function baseCreateLanguageWorker(host, vueCompilerOptions, ts) {
  const vueLanguages = ts ? (
    // eslint-disable-next-line etc/no-deprecated
    vue.createLanguages(host.getCompilationSettings(), vueCompilerOptions, ts)
  ) : [];
  const core = vue.createLanguageContext(host, vueLanguages);
  const tsLsHost = createLanguageServiceHost(core, ts, ts.sys);
  const tsLs = ts.createLanguageService(tsLsHost);
  decorateLanguageService(core.virtualFiles, tsLs, false);
  const getScriptKind = tsLsHost.getScriptKind;
  tsLsHost.getScriptKind = (fileName) => {
    if (fileName.endsWith(".vue.js")) {
      return ts.ScriptKind.TS;
    }
    if (fileName.endsWith(".vue.jsx")) {
      return ts.ScriptKind.TSX;
    }
    return getScriptKind(fileName);
  };
  const program = tsLs.getProgram();
  const typeChecker = program.getTypeChecker();
  function findScriptRangesNode(filepath, filter) {
    filepath = normalizePath(filepath);
    const sourceFile = core.virtualFiles.getSource(filepath)?.root;
    if (!(sourceFile instanceof vue.VueFile)) {
      return;
    }
    if (!sourceFile.sfc.scriptSetup) {
      return;
    }
    const {
      ast: scriptSetupAst,
      startTagEnd,
      lang
    } = sourceFile.sfc.scriptSetup;
    if (lang !== "ts" && lang !== "tsx") {
      return;
    }
    const virtualTsFile = program.getSourceFile(`${filepath}.${lang}`);
    if (!virtualTsFile) {
      return;
    }
    const scriptSetupRanges = vue.parseScriptSetupRanges(
      ts,
      scriptSetupAst,
      vueCompilerOptions
    );
    const virtualTsFileScriptSetupRanges = vue.parseScriptSetupRanges(
      ts,
      virtualTsFile,
      vueCompilerOptions
    );
    const sourceRange = filter(scriptSetupRanges);
    const virtualTsRange = filter(virtualTsFileScriptSetupRanges);
    if (!sourceRange || !virtualTsRange) {
      return;
    }
    let foundNode;
    virtualTsFile.forEachChild(function traverse(node) {
      if (node.getStart(virtualTsFile) === virtualTsRange.start && node.getEnd() === virtualTsRange.end) {
        foundNode = node;
      } else {
        node.forEachChild(traverse);
      }
    });
    const range = {
      start: startTagEnd + sourceRange.start,
      end: startTagEnd + sourceRange.end
    };
    return { node: foundNode, range };
  }
  return {
    findScriptRangesNode,
    __internal__: {
      tsLs,
      program,
      typeChecker
    }
  };
}
const LANGUAGE_GLOBAL_KEY = "__UNPLUGIN_VUE_COMPLEX_LANGUAGE__";
function ensureLanguage(tsconfigPath) {
  if (!globalThis[LANGUAGE_GLOBAL_KEY]) {
    globalThis[LANGUAGE_GLOBAL_KEY] = createLanguage(tsconfigPath);
  }
}
function getLanguage() {
  if (!globalThis[LANGUAGE_GLOBAL_KEY]) {
    throw new Error("[unplugin-vue-complex-types] Language not created!");
  }
  return globalThis[LANGUAGE_GLOBAL_KEY];
}

class Printer {
  constructor(checker) {
    this.checker = checker;
  }
  getBaseType(nodeOrType) {
    if ("kind" in nodeOrType) {
      nodeOrType = this.checker.getTypeAtLocation(nodeOrType);
    }
    return this.checker.getBaseConstraintOfType(nodeOrType) ?? this.checker.getBaseTypeOfLiteralType(nodeOrType);
  }
  printIntersectionTypeNode(node) {
    return node.types.map((t) => this.print(t)).join(" & ");
  }
  printUnionTypeNode(node) {
    return node.types.map((t) => this.print(t)).join(" | ");
  }
  printTypeLiteralNode(node) {
    const parts = ["{"];
    for (const member of node.members) {
      if (ts.isPropertySignature(member)) {
        const stringBaseType = member.type ? this.checker.typeToString(this.getBaseType(member.type)) : "any";
        parts.push(
          [
            member.name.getText(),
            member.questionToken?.getText(),
            ": ",
            stringBaseType
          ].filter(Boolean).join("")
        );
      }
    }
    parts.push("}");
    return parts.join("\n");
  }
  printByType(node) {
    const type = this.checker.getTypeAtLocation(node);
    const properties = type.getProperties();
    const parts = ["{"];
    let questionToken = "";
    if (ts.isMappedTypeNode(node)) {
      questionToken = node.questionToken?.getText() ?? "";
    }
    for (const property of properties) {
      const valueType = this.checker.getTypeOfSymbol(property);
      const stringValueType = this.checker.typeToString(
        this.getBaseType(valueType)
      );
      parts.push(
        `${this.checker.symbolToString(
          property
        )}${questionToken}: ${stringValueType}`
      );
    }
    parts.push("}");
    return parts.join("\n");
  }
  print(node) {
    if (ts.isIntersectionTypeNode(node)) {
      return this.printIntersectionTypeNode(node);
    } else if (ts.isUnionTypeNode(node)) {
      return this.printUnionTypeNode(node);
    } else if (ts.isTypeLiteralNode(node)) {
      return this.printTypeLiteralNode(node);
    } else if (ts.isMappedTypeNode(node) || ts.isTypeReferenceNode(node)) {
      return this.printByType(node);
    } else {
      console.error(
        `[unplugin-vue-complex-types] \`${ts.SyntaxKind[node.kind]}\` is not supported.`
      );
      return node.getText();
    }
  }
}

function transform(code, id) {
  const language = getLanguage();
  const s = new MagicString(code);
  const typeChecker = language.__internal__.typeChecker;
  const printer = new Printer(typeChecker);
  const definePropsTypeArg = language.findScriptRangesNode(
    id,
    (scriptSetupRanges) => scriptSetupRanges.props.define?.typeArg
  );
  if (!definePropsTypeArg) {
    return;
  }
  const { node: typeArgNode, range: typeArgRange } = definePropsTypeArg;
  const printedType = printer.print(typeArgNode);
  s.overwrite(typeArgRange.start, typeArgRange.end, printedType);
  return {
    code: s.toString(),
    map: s.generateMap({ hires: true })
  };
}

const unplugin = createUnplugin((options = {}) => ({
  name: "unplugin-vue-complex-types",
  buildStart() {
    const resolvedOptions = resolveOptions(options);
    ensureLanguage(resolvedOptions.tsconfigPath);
  },
  transform(code, id) {
    if (!id.endsWith(".vue")) {
      return;
    }
    return transform(code, id);
  }
}));

export { unplugin as default, ensureLanguage, getLanguage };
